import logging
from random import random

from pyage.core.address import Addressable
from pyage.core.inject import Inject
from pyage_forams.solutions.genom import Genom
from pyage_forams.utils import counted


logger = logging.getLogger(__name__)


class ForamAggregateAgent(Addressable):
    @Inject("forams", "environment")
    def __init__(self):
        super(ForamAggregateAgent, self).__init__()
        for foram in self.forams.values():
            foram.parent = self
            self.environment.add_foram(foram)

    def step(self):
        for foram in self.forams.values():
            foram.step()
        self.environment.tick(self.parent.steps)

    def remove_foram(self, address):
        foram = self.forams[address]
        del self.forams[address]
        foram.parent = None
        return foram

    def add_foram(self, foram):
        foram.parent = self
        self.forams[foram.get_address()] = foram


class Foram(Addressable):
    @Inject("genom_factory", "reproduction_minimum", "movement_energy", "growth_minimum", "energy_need")
    def __init__(self, energy, genom=None):
        super(Foram, self).__init__()
        self.energy = energy
        if genom is None:
            self.genom = self.genom_factory()
        else:
            self.genom = genom
        self.steps = 0
        self.chambers = 0
        self.alive = True
        self.cell = None

    def step(self):
        if not self.alive:
            logger.warn("called step on dead foram")
            return
        logger.debug("foram! %s" % self.steps)
        self.steps += 1
        if self._should_die():
            self._die()
            return
        if self._eat() <= 0:
            self._move()
        if self._can_reproduce():
            self._reproduce()
        if self._can_create_chamber():
            self._create_chamber()

    def _eat(self):
        e = self.energy
        capacity = self._energy_capacity()
        self.energy += self._take_algae(capacity) - self._energy_demand()
        return self.energy - e

    def _energy_capacity(self):
        capacity = self.chambers + 1
        return capacity

    def _energy_demand(self):
        return self.energy_need * (self.chambers + 1)

    def _take_algae(self, capacity):
        taken = 0
        cells = iter([self.cell] + self.cell.neighbours)
        while capacity > taken:
            try:
                cell = cells.next()
                taken += cell.take_algae(capacity - taken)
            except StopIteration:
                break
        return taken

    def _can_reproduce(self):
        return self.energy > self.reproduction_minimum and self.genom.chambers_limit <= self.chambers

    @counted
    def _reproduce(self):
        logger.debug("%s is reproducing" % self)
        empty_neighbours = filter(lambda c: c.is_empty(), self.cell.neighbours)
        if not empty_neighbours:
            return
        energy = self.energy / (len(empty_neighbours) * 2.0)
        for cell in empty_neighbours:
            self.energy -= energy
            foram = Foram(energy, Genom(self.genom.chambers_limit))
            cell.insert_foram(foram)
            self.parent.add_foram(foram)

    def _move(self):
        try:
            empty_neighbours = filter(lambda c: c.is_empty(), self.cell.neighbours)
            logger.debug(self.cell.neighbours)
            logger.debug(empty_neighbours)
            cell = max(empty_neighbours,
                       key=lambda c: random() + c.available_food() + sum(s.available_food() for s in c.neighbours))
            if cell:
                cell.insert_foram(self.cell.remove_foram())
                self.energy -= self.movement_energy
                logger.debug("%s moved" % self)
        except:
            logging.exception("could not move")

    def _should_die(self):
        return self.energy <= 0

    def _die(self):
        logger.debug("%s died" % self)
        self.alive = False
        self.parent.remove_foram(self.get_address())
        self.cell.remove_foram()

    def _can_create_chamber(self):
        return self.energy > self.growth_minimum and self.genom.chambers_limit > self.chambers

    def _create_chamber(self):
        self.energy -= 5
        self.chambers += 1
        logger.debug("Foram %s has a new chamber, so %d altogether" % (self, self.chambers))

    def __repr__(self):
        return "Foram[%.2f, %d]" % (self.energy, self.chambers)


def create_forams(count, initial_energy):
    def factory():
        forams = {}
        for i in range(count):
            foram = Foram(initial_energy)
            forams[foram.get_address()] = foram
        return forams

    return factory


def create_agent():
    agent = ForamAggregateAgent()
    return {agent.get_address(): agent}


# coding=utf-8

from pyage.core import address

from pyage.core.stop_condition import StepLimitStopCondition
from pyage_forams.solutions.environment import environment_factory, Environment3d
from pyage_forams.solutions.foram import create_forams, create_agent
from pyage_forams.solutions.genom import GenomFactory
from pyage_forams.solutions.insolation_meter import StaticInsolation
from pyage_forams.solutions.statistics import SimpleStatistics


factory = GenomFactory(chambers_limit=5)
genom_factory = lambda: factory.generate
forams = create_forams(8, initial_energy=5)
agents = create_agent
insolation_meter = StaticInsolation
size = lambda: 10

reproduction_minimum = lambda: 10
movement_energy = lambda: 0.25
growth_minimum = lambda: 10
energy_need = lambda: 0.2

environment = environment_factory(regeneration_factor=0.1, clazz=Environment3d)

stop_condition = lambda: StepLimitStopCondition(90)

address_provider = address.SequenceAddressProvider
stats = SimpleStatistics


# coding=utf-8
from paver.easy import *
from paver.setuputils import setup
from setuptools import find_packages

setup(
    name="pyage-forams",
    description="Forams package for Pyage platform",
    packages=find_packages(),
    version="0.6.0",
    author="Maciej Kaziród",
    author_email="kazirod.maciej@gmail.com",
    requires=['matplotlib', 'pyage']
)


@task
@needs('generate_setup', 'minilib', 'setuptools.command.sdist')
def sdist():
    """Overrides sdist to make sure that our setup.py is generated."""
    pass


@task
@needs("setuptools.command.bdist_egg")
def egg():
    pass



from random import randint, choice, random

from pyage.core.inject import Inject


class AbstractEnvironment(object):
    @Inject("insolation_meter")
    def __init__(self, regeneration_factor):
        super(AbstractEnvironment, self).__init__()
        self.regeneration_factor = regeneration_factor

    def add_foram(self, foram):
        choice(filter(lambda c: c.is_empty(), self.get_all_cells())).insert_foram(foram)

    def tick(self, step):
        for cell in self.get_all_cells():
            if cell.algae > 0:
                cell.algae += self.regeneration_factor + self.insolation_meter.get_insolation(cell, step)
        while random() > 0.4:
            try:
                choice(filter(lambda c: c.is_empty(), self.get_all_cells())).algae += 1 + random() * 2
            except:
                pass


class Environment2d(AbstractEnvironment):
    @Inject("size")
    def __init__(self, regeneration_factor):
        super(Environment2d, self).__init__(regeneration_factor)
        self.grid = self._initialize_grid()

    def _initialize_grid(self):
        grid = [[Cell(randint(0, 4) if random() > 0.7 else 0) for _ in range(self.size)] for _ in range(self.size)]
        for i in range(self.size):
            for j in range(self.size):
                grid[i][j].neighbours.extend([grid[x][y] for x in range(max(0, i - 1), min(self.size, i + 2)) for y in
                                              range(max(0, j - 1), min(self.size, j + 2)) if x != i or y != j])

        return grid

    def get_all_cells(self):
        for row in self.grid:
            for cell in row:
                yield cell


class Environment3d(AbstractEnvironment):
    @Inject("size")
    def __init__(self, regeneration_factor):
        super(Environment3d, self).__init__(regeneration_factor)
        self.grid = self._initialize_grid()

    def _initialize_grid(self):
        grid = [[[Cell(randint(1, 4) if random() > 0.7 else 0)
                  for _ in range(self.size)] for _ in range(self.size)] for _ in range(self.size)]
        for i in range(self.size):
            for j in range(self.size):
                for k in range(self.size):
                    grid[i][j][k].depth = self.size - k - 1
                    grid[i][j][k].neighbours.extend(
                        [grid[x][y][z] for x in range(max(0, i - 1), min(self.size, i + 2)) for y in
                         range(max(0, j - 1), min(self.size, j + 2)) for z in
                         range(max(0, k - 1), min(self.size, k + 2)) if x != i or y != j or z != k])

        return grid

    def get_all_cells(self):
        for plane in self.grid:
            for row in plane:
                for cell in row:
                    yield cell


class Cell(object):
    def __init__(self, algae=0):
        super(Cell, self).__init__()
        self.algae = algae
        self.foram = None
        self.neighbours = []

    def insert_foram(self, foram):
        if not self.is_empty():
            raise ValueError("cannot insert foram to already occupied cell")
        self.foram = foram
        foram.cell = self

    def remove_foram(self):
        foram = self.foram
        self.foram = None
        return foram

    def is_empty(self):
        return self.foram is None

    def take_algae(self, demand):
        to_let = min(demand, self.algae)
        self.algae -= to_let
        return to_let

    def available_food(self):
        return self.algae + sum(map(lambda c: c.algae, self.neighbours))

    def __repr__(self):
        return "%d, %s" % (self.algae, self.foram)


def environment_factory(regeneration_factor=0.1, clazz=Environment2d):
    e = [None]

    def environ():
        if e[0] is None:
            e[0] = clazz(regeneration_factor)
        return e[0]

    return environ


from unittest import TestCase
from pyage.core import inject
from pyage_forams.solutions.environment import Environment2d, Environment3d


class TestEnvironment(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestEnvironment, cls).setUpClass()
        inject.config = "pyage_forams.conf.dummy_conf"

    def test_environment(self):
        self._get_environments()

    def test_neighborhood(self):
        for environment in self._get_environments():
            environment.get_all_cells()

    def test_neighborhood_symmetry(self):
        for environment in self._get_environments():
            for cell in environment.get_all_cells():
                for neighbour in cell.neighbours:
                    self.assertTrue(cell in neighbour.neighbours)

    def test_self_neighbourhood(self):
        for environment in self._get_environments():
            for cell in environment.get_all_cells():
                self.assertFalse(cell in cell.neighbours)

    def test_neighbourhood_repetitions(self):
        for environment in self._get_environments():
            for cell in environment.get_all_cells():
                self.assertEqual(len(cell.neighbours), len(set(cell.neighbours)))

    def test_2d_neighbourhood(self):
        grid = (self._get_environments()[0]).grid
        self.assertEqual(len(grid[1][1].neighbours), 8)

        self.assertTrue(grid[0][0] in grid[1][1].neighbours)
        self.assertTrue(grid[0][1] in grid[1][1].neighbours)
        self.assertTrue(grid[0][2] in grid[1][1].neighbours)
        self.assertTrue(grid[1][0] in grid[1][1].neighbours)
        self.assertTrue(grid[1][2] in grid[1][1].neighbours)
        self.assertTrue(grid[2][0] in grid[1][1].neighbours)
        self.assertTrue(grid[2][1] in grid[1][1].neighbours)
        self.assertTrue(grid[2][2] in grid[1][1].neighbours)

    def test_3d_neighbourhood(self):
        grid = (self._get_environments()[1]).grid
        self.assertEqual(len(grid[1][1][1].neighbours), 26)

        self.assertTrue(grid[0][0][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][0][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][0][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][0][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][1][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][1][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][1][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][2][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][2][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[0][2][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][0][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][0][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][0][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][1][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][1][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][2][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][2][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[1][2][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][0][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][0][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][0][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][0][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][0][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][0][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][1][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][1][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][1][2] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][2][0] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][2][1] in grid[1][1][1].neighbours)
        self.assertTrue(grid[2][2][2] in grid[1][1][1].neighbours)

    def _get_environments(self):
        return [Environment2d(0.2), Environment3d(0.2)]


from __future__ import print_function
from compiler.ast import flatten
from datetime import datetime
import itertools
import logging
from subprocess import Popen

import matplotlib.pyplot as plt

from pyage.core.inject import Inject
from pyage.core.statistics import Statistics
from pyage_forams.solutions.foram import Foram


logger = logging.getLogger(__name__)


class PlottingStatistics(Statistics):
    @Inject("environment")
    def __init__(self):
        super(PlottingStatistics, self).__init__()
        self.prefix = datetime.now().strftime("%Y%m%d_%H%M%S")

    def update(self, step_count, agents):
        self.plot(self.environment.grid, step_count)

    def summarize(self, agents):
        cmd = "convert -loop 0 -delay 50 /tmp/pyage-%s-step-*.png animation.gif && rm /tmp/pyage-%s-step-*.png" % (
            self.prefix, self.prefix)
        Popen(cmd, shell=True)

    def plot(self, grid, step):
        plt.title("step %d" % step)

        forams = []
        for (i, row) in enumerate(grid):
            for (j, cell) in enumerate(row):
                if not cell.is_empty():
                    forams.append((i, j))
                    self._draw_chambers(grid, i, j)

        x = len(grid)
        y = len(grid[0])

        self._draw_forams(forams, grid)
        self._draw_algae(grid, x, y)
        self._draw_grid(x, y)

        plt.savefig("/tmp/pyage-%s-step-%07d.png" % (self.prefix, step))
        plt.close()

    @staticmethod
    def _draw_forams(forams, grid):
        plt.scatter([c[0] for c in forams],
                    [c[1] for c in forams],
                    marker='o',
                    s=[200.0 / len(grid) * grid[i][j].foram.energy for (i, j) in forams],
                    c=['red' for _ in forams])

    @staticmethod
    def _draw_grid(x, y):
        plt.xticks([i - 0.5 for i in range(x + 1)], [])
        plt.yticks([i - 0.5 for i in range(y + 1)], [])
        plt.ticklabel_format()
        plt.xlim([-0.5, x - 0.5])
        plt.ylim([-0.5, y - 0.5])
        plt.grid()

    @staticmethod
    def _draw_algae(grid, x, y):
        pts = itertools.product(range(x), range(y))
        plt.scatter(*zip(*pts), marker='*',
                    s=[1000.0 / x * grid[i][j].algae for i in range(x) for j in range(y)], color='blue',
                    alpha=0.5)

    @staticmethod
    def _draw_chambers(grid, i, j):
        plt.annotate(
            grid[i][j].foram.chambers,
            xy=(i, j), xytext=(-5, 5),
            textcoords='offset points', ha='right', va='bottom',
            bbox=dict(boxstyle='round,pad=0.2', fc='yellow', alpha=0.5),
            arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))


class CsvStatistics(Statistics):
    @Inject("environment")
    def __init__(self, interval=1, filename=None):
        super(CsvStatistics, self).__init__()
        self.interval = interval
        self.filename = "forams-%s.log" % datetime.now().strftime("%Y%m%d_%H%M%S") if filename is None else filename

    def update(self, step_count, agents):
        if step_count % self.interval == 0:
            with open(self.filename, 'ab') as f:
                entry = self._get_entry(agents, step_count)
                print((",".join(map(str, entry))), file=f)

    def summarize(self, agents):
        pass

    def _get_entry(self, agents, step_count):
        forams_count = len(agents[0].forams)
        entry = [step_count, forams_count,
                 sum(f.chambers for f in agents[0].forams.values()) / float(forams_count) if forams_count > 0 else 0,
                 sum(c.algae for row in self.environment.grid for c in row),
                 Foram._reproduce.called]
        return entry


class PsiStatistics(Statistics):
    @Inject("environment", "insolation_meter", "stop_condition")
    def __init__(self, interval=1, filename=None):
        super(PsiStatistics, self).__init__()
        self.interval = interval
        self._column_names = ['"x"', '"y"', '"z"', '"Foram"', '"Algae"', '"Insolation"']
        self._column_symbols = ['"F"', '"A"', '"I"']
        self._column_types = ['float', 'float', 'float']
        self.filename = "forams-%s" % datetime.now().strftime("%Y%m%d_%H%M%S") if filename is None else filename
        self.counter = 0

    def update(self, step_count, agents):
        if step_count % self.interval == 0:
            self.counter += 1
            new_filename = '%s%s.psi' % (self.filename, '%06d' % self.counter)
            with open(new_filename, 'w') as f:
                self._add_header(f)
                self._add_data(f)

    def summarize(self, agents):
        pass

    def _add_header(self, f):
        f.write('# PSI Format 1.0\n#\n')
        self._add_column_names(f)
        self._add_column_symbols(f)
        self._add_column_types(f)
        f.write('%d 2694 115001\n'
                     '1.00 0.00 0.00\n'
                     '0.00 1.00 0.00\n'
                     '0.00 0.00 1.00\n\n'
                     % len(flatten(self.environment.grid)))

    def _add_data(self, f):
        for x in range(len(self.environment.grid)):
            for y in range(len(self.environment.grid[x])):
                for z in range(len(self.environment.grid[x][y])):
                    f.write(' '.join(map(str, self._get_entry(x, y, z))) + '\n')

    def _get_entry(self, x, y, z):
        cell = self.environment.grid[x][y][z]
        return map(float, [x, y, z] + [0 if cell.is_empty() else cell.foram.chambers,
                                       cell.algae,
                                       self.insolation_meter.get_insolation(cell)])

    def _add_column_names(self, f):
        names = (['# column[%d] = %s' % (i, n) for i, n in enumerate(self._column_names)])
        f.write('\n'.join(names) + '\n')

    def _add_column_symbols(self, f):
        symbols = ['# symbol[%d] = %s' % (i + 3, s) for i, s in enumerate(self._column_symbols)]
        f.write('\n'.join(symbols) + '\n')

    def _add_column_types(self, f):
        types = ['# type[%d] = %s' % (i + 3, t) for i, t in enumerate(self._column_types)]
        f.write('\n'.join(types) + '\n')


class SimpleStatistics(Statistics):
    @Inject("environment", "insolation_meter")
    def __init__(self):
        super(SimpleStatistics, self).__init__()

    def update(self, step_count, agents):
        if step_count % 15 == 0:
            logger.info(self.environment.grid)

    def summarize(self, agents):
        logger.debug("done")

